#include "crogue.h"

int high_density_font=0;
int monochrome_screen=0;
int game_running = 0;
int game_repeat = 0;

//{{{
UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags)
{
	if(cmd != sysAppLaunchCmdNormalLaunch)
		return 0;
	
	do
	{
		game_repeat = 0;
		
		ErrTry
		{
			initialize();
			eventloop();
		}
		ErrCatch(inErr)
		{
			(void)inErr;
		} ErrEndCatch
	} while(game_repeat);

	return 0;
}
//}}}
//{{{
void display_title_screen(void)
{
	MemHandle bitmap_handle = DmGetResource('Tbmp', bitmap_splash);
	BitmapType *bitmap = MemHandleLock(bitmap_handle);
	WinDrawBitmap(bitmap, 0, 0);
	MemHandleUnlock(bitmap_handle);
}
//}}}

int color_black, color_dgray, color_lgray, color_white;
MemHandle data_file_handle;
//{{{
void sys_init(void)
{
	Err error;
	UInt32 density;
	unsigned long version;
	UInt32 winVersion;
	MemHandle fontHandle;
	RGBColorType rgb;
	FontType *fontPtr = NULL;
	
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &version);
	
	// Load itsy font
	error = FtrGet(sysFtrCreator, sysFtrNumWinVersion, &winVersion);
	if(!error && winVersion>=4) {
		error = WinScreenGetAttribute(winScreenDensity, &density);
		if(!error && density != kDensityLow) {
			fontHandle = DmGetResource('nfnt', FONT_8_12);
			fontPtr = MemHandleLock(fontHandle);
			FntDefineFont(TextFont, fontPtr);
			high_density_font = 1;
		}
	}
	if(!fontPtr)
	{
		if (version >= 0x03000000L) {
			fontHandle = DmGetResource('NFNT', FONT_5_6);
			fontPtr = MemHandleLock(fontHandle);
			FntDefineFont(TextFont, fontPtr);
			high_density_font = 0;
		}
	}
	FntSetFont(TextFont);
	
	// Load tile font
	fontHandle = DmGetResource('NFNT', TILE_5_5);
	fontPtr = MemHandleLock(fontHandle);
	FntDefineFont(TileFont, fontPtr);
	
	if(version >= 0x3503000) // PalmOS 3.5 with color support
	{
		// Set up colors
		rgb.r=rgb.g=rgb.b=255;
		color_white = WinRGBToIndex(&rgb);
		rgb.r=rgb.g=rgb.b=171;
		color_lgray = WinRGBToIndex(&rgb);
		rgb.r=rgb.g=rgb.b=86;
		color_dgray = WinRGBToIndex(&rgb);
		rgb.r=rgb.g=rgb.b=0;
		color_black = WinRGBToIndex(&rgb);
		
		if(color_black == color_dgray) // Suspiciously few colors...
			monochrome_screen = 1;
	}
	else
	{
		monochrome_screen = 1;
	}
	
	FrmGotoForm(MainForm);
	handle_events();
}
//}}}
//{{{
void sys_cleanup(void)
{
	MemHandle fontHandle;
	if(high_density_font)
		fontHandle = DmGetResource('nfnt', FONT_8_12);
	else
		fontHandle = DmGetResource('NFNT', FONT_5_6);
	MemHandleUnlock(fontHandle);
	
	fontHandle = DmGetResource('NFNT', TILE_5_5);
	MemHandleUnlock(fontHandle);
	
	if(w->constfileoffset) {
//		debug_free(w->constfileoffset);
		MemHandleUnlock(data_file_handle);
		DmReleaseResource(data_file_handle);
		w->constfileoffset = NULL;
	}
	FrmCloseAllForms();
}
//}}}
//{{{
void init_layout(void)
{
	uint tile_width = 5,
	     tile_height = 5;
	uint message_rows = 3;
	const char *tileset = tiles;
	
	int screen_top = 0;
	int screen_bottom = SCREEN_HEIGHT;
	
	screen_bottom -= 10; // Make room for bottonbar
	
	UI_MF_Place(0, message_rows);
	screen_top += message_rows * FontHeight();
	
	screen_bottom -= FontHeight();
	UI_Status_Place(screen_bottom, FontHeight());
	
	UI_TF_Place( COORD(0, screen_top),
	             COORD(SCREEN_WIDTH, screen_bottom-screen_top-1),
	             tileset, tile_height, tile_width );
}
//}}}
//{{{
void sys_init_data_file(void)
{
/*	FILE *constfile;
	
	// Load constant-data file
	constfile = fopen("crogdat", "rb");
	if(!constfile)
		panic("Couldn't open constant data file.");
	/// @todo FIXME Gigantic whopping awful unnecessary waste of heap!
	w->constfileoffset = debug_malloc(64000);
	if(!w->constfileoffset)
		panic("Allocation failed!");
	if( fread(w->constfileoffset, 1, 64000, constfile) == 0 )
		panic("Loaded nothing!");
	memcpy(&w->desc, w->constfileoffset, sizeof(struct world_descriptor));
	fclose(constfile);*/
	data_file_handle = DmGetResource('CRdt', DataFile);
	if(!data_file_handle) panic("Couldn't fetch data file");
	w->constfileoffset = MemHandleLock(data_file_handle);
	if(!w->constfileoffset) panic("Couldn't lock data file");
	memcpy(&w->desc, w->constfileoffset, sizeof(struct world_descriptor));
	
	link_data_file();
}
//}}}

//{{{
void PalmExit(int ret)
{
	ErrThrow(0);
}
//}}}

