#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <stdio.h>
#include <string.h>
using namespace std;

bool isLittleEndian=false;
//{{{
void reverseBytes(char *target, int num_bytes)
{
	if(isLittleEndian)
	{
		int i;
		char *buffer = new char[num_bytes];
		
		for(i=0; i<num_bytes; i++)
			buffer[i] = target[i];
		for(i=0; i<num_bytes; i++)
			target[i] = buffer[num_bytes-i-1];
	}
}
//}}}
//{{{
/**/template<class T>
class Endian
{
public:
	Endian() {}
	Endian(T d) {
		data = d;
		reverseBytes((char*)(&data), sizeof(T));
	}
	Endian<T> operator= (const T d) {
		data = d;
		reverseBytes((char*)(&data), sizeof(T));
		return *this;
	}
	operator T() {
		T ret = data;
		reverseBytes((char*)(&ret), sizeof(T));
		return ret;
	}
	istream& operator>>(istream& stream) {
		int x=data;
		stream>>x;
		return stream;
	}
	ostream& operator<<(ostream& stream) {
		int x;
		stream<<x;
		data = x;
		return stream;
	}
	
private:
	T data;
};
//}}}

//{{{
#define ulong  Endian<unsigned long>
#define slong  Endian<signed   long>
#define uint   Endian<unsigned short>
#define sint   Endian<signed   short>
#define ushort Endian<unsigned short>
#define sshort Endian<signed   short>
#define uchar  Endian<unsigned char>
#define schar  Endian<signed   char>
//}}}

//{{{
class Error
{
public:
	virtual ~Error() {}
	virtual void write(ostream& stream) {}
};
ostream& operator<<(ostream& stream, Error& err)
{
	err.write(stream);
	return stream;
}
//}}}
//{{{
class StringError : public Error
{
public:
	StringError(string p_str) {
		str = p_str;
	}
	~StringError() {}
	void write(ostream& stream) {
		stream<<str;
	}
private:
	string str;
};
//}}}
//{{{
class ArrayBoundsError : public Error
{
public:
	ArrayBoundsError(int p_index, int p_size) {
		index = p_index;
		size = p_size;
	}
	
	void write(ostream& stream) {
		stream<<"Internal error: Out of array bounds. Tried "<<index<<" on size "<<size;
	}
	
private:
	int index, size;
};
//}}}
//{{{
/**/template<class T>
class Array
{
public:
	Array() {
		size = 0;
		data = new T[1];
	}
	Array(int p_size) {
		size = p_size;
		data = new T[p_size];
	}
	
	T& operator[](int index) {
		if(index<0 || index>=size)
			throw ArrayBoundsError(index, size);
		
		return data[index];
	}
	T& operator*() {
		return operator[](0);
	}
	T& end(void) {
		return operator[](size-1);
	}
	
	void resize(int p_size) {
		int copyEntries = p_size;
		if(copyEntries>size)
			copyEntries = size;
		
		T *oldData = data;
		data = new T[p_size];
		
		for(int i=0; i<copyEntries; i++)
			data[i] = oldData[i];
		
		delete[] oldData;
		size = p_size;
	}
	int getSize(void) {
		return size;
	}
	
	void append(Array<T>& oth)
	{
		int startPos = size;
		resize(size+oth->getSize());
		
		for(int i=startPos; i<size; i++)
		{
			data[i] = oth[i];
		}
	}
	void append(T& oth)
	{
		resize(size+1);
		data[size-1] = oth;
	}
	
private:
	int size;
	T *data;
};
//}}}

//{{{
enum token_type
{
	token_invalid = 0,
	token_regular,
	token_quotedstring,
	token_quotedchar,
	token_numericConstant,
	token_group,
	token_parenthesized,
	token_misc
};
//}}}
//{{{
bool iswhite(char c)
{
	if(c==' '||c=='\t'||c=='\n')
		return true;
	else
		return false;
}
//}}}
//{{{
bool weof(istream& stream)
{
	ws(stream);
	return stream.eof();
}
//}}}
//{{{
string intToString(int x)
{
	char buf[16];
	sprintf(buf, "%i", x);
	return buf;
}
//}}}
//{{{
int stringToInt(string x)
{
	char strbuf[16];
	strcpy(strbuf, x.c_str());
	int buf;
	sscanf(strbuf, "%d", &buf);
	return buf;
}
//}}}
//{{{
string readToken(istream& stream, token_type &type)
{
	char next;
	int writePos = 0;
	int i;
	char buffer[4096];
	
	if(stream.bad()) throw StringError("Internal error: Input stream is marked BAD.");
	if(stream.eof()) throw StringError("Unexpected end of file found.");
	
	ws(stream);
	next = stream.peek();
	if( isalpha(next) || next=='_' ) // Regular tokens
	{
		type = token_regular;
		do
		{
			buffer[writePos++] = stream.get();
			next = stream.peek();
		} while( !stream.eof() && (isalpha(next) || isdigit(next) || next=='_') );
	}
	else if( next=='\"' ) // Quoted strings come with surrounding quotes removed, all together
	{
		type = token_quotedstring;
		stream.get();
		next = stream.peek();
		while( next != '\"' )
		{
			buffer[writePos++] = stream.get();
			if(stream.eof())
				throw StringError("Unexpected end of file found. Mismatched quotes.");
			next = stream.peek();
		}
		stream.get();
	}
	else if( next=='<' ) // Used to be 'c', now it's <c> (to avoid conflicts with m4's quoting mechanism)
	{
		type = token_quotedchar;
		stream.get();
		if(stream.peek()=='\\') stream.get();
		i = stream.get();
		sprintf(buffer, "%i", i);
		
		if(stream.get() != '>')
			throw StringError("Character constant too long.");
		return buffer;
	}
	else if( isdigit(next) || next=='-' ) // Numeric constants
	{
		type = token_numericConstant;
		stream>>i;
		sprintf(buffer, "%i", i);
		return buffer;
	}
	else if( next=='(' ) // Indirection. (N) refers to the Nth entry in the shuffle table.
	{
		type = token_parenthesized;
		stream.get();
		stream >> i;
		ws(stream);
		stream.get();
		sprintf(buffer, "(%i)", i);
		return buffer;
	}
	else // For every thing else, f* it and return one char
	{
		type = token_misc;
		buffer[writePos++] = stream.get();
	}
	
	buffer[writePos] = '\0';
	return buffer;
}
//}}}

// FIXME: this is duplication of typing data
struct filelink
{
	uchar flags;
	uchar file;
	uint offset;
};
#define PTRFLAG_JUMPTABLE  1
#define PTRFLAG_SHUFFLETAB 2

//{{{
class Linker
{
public:
	struct linkEntry
	{
		linkEntry() {
		}
		linkEntry(string p_name, int p_position, bool p_isTable) {
			name = p_name;
			position = p_position;
			isTable = p_isTable;
		}
		string name;
		int position;
		bool isTable;
	};
	
	Linker()
	{
		fout = fopen("gamedat.out", "wb");
		if(!fout)
			throw StringError("Couldn't open output file for writing.");
	}
	void reset(void)
	{
		position = 0;
	}
	
	void writeExport( string linkerID, bool isTable )
	{
		linkEntry entry(linkerID, position, isTable);
		linktable.append( entry );
	}
	void writeData( bool dryRun, size_t size, const void *data )
	{
		if(dryRun)
			position += size;
		else
		{
			fwrite(data, 1, size, fout);
			position += size;
		}
	}
	void align(bool dryRun) // If on an odd address, skip a byte (for things that must be on even addresses)
	{
		if( position%2 )
		{
			int pad=0;
			writeData( dryRun, 1, &pad );
		}
	}
	filelink findSymbol(string symbol)
	{
		filelink ret;
		
		if(symbol[0] == '(') // Indirect link; points to shuffletab
		{
			int t;
			sscanf(symbol.c_str(), "(%i)", &t);
			
			ret.file = 0;
			ret.flags = PTRFLAG_SHUFFLETAB;
			ret.offset = t;
			return ret;
		}
		
		for(int i=0; i<linktable.getSize(); i++)
		{
			if( !linktable[i].name.compare(symbol) ) {
				ret.file = 0;
				if(linktable[i].isTable)
					ret.flags = PTRFLAG_JUMPTABLE;
				else
					ret.flags = 0;
				ret.offset = linktable[i].position;
				return ret;
			}
		}
		string errorMessage = "Undefined symbol: ";
		errorMessage.append(symbol);
		errorMessage.append(".");
		throw StringError(errorMessage);
	}
	
private:
	FILE *fout;
	int position;
	Array<linkEntry> linktable;
};
//}}}
Linker linker;
//{{{
class Expression
{
public:
	virtual ~Expression() {}
	virtual void create(string p_type, string contents)=0;
	virtual void evaluate(void)=0;
	virtual void write(bool dryRun)=0;
	virtual int  size(void)=0;
	
protected:
	string name;
};
//}}}
//{{{
/**/template<class T>
class ExpressionNumeric : public Expression
{
public:
	void create(string p_type, string contents)
	{
		data = (T)stringToInt(contents);
	}
	void evaluate(void)
	{
	}
	void write(bool dryRun)
	{
		if(sizeof(data) > 1)
			linker.align(dryRun);
		linker.writeData( dryRun, sizeof(data), &data );
	}
	int  size(void)
	{
		return sizeof(T);
	}

private:
	T data;
};
//}}}
//{{{
class ExpressionExport : public Expression
{
public:
	void create(string p_type, string contents)
	{
		symbol = contents;
	}
	virtual void evaluate(void)
	{
	}
	virtual void write(bool dryRun)
	{
		if(dryRun)
			linker.writeExport(symbol, isTable());
	}
	virtual int  size(void)
	{
		return 0;
	}
protected:
	virtual bool isTable() { return false; }
	string symbol;
};
//}}}
//{{{
class ExpressionExportTable : public ExpressionExport
{
protected:
	bool isTable() { return true; }
};
//}}}
//{{{
class ExpressionLink : public Expression
{
public:
	void create(string p_type, string contents)
	{
		symbol = contents;
	}
	void evaluate(void)
	{
		data = linker.findSymbol(symbol);
	}
	void write(bool dryRun)
	{
		linker.align(dryRun);
		linker.writeData( dryRun, sizeof(data), &data );
	}
	int  size(void)
	{
		return sizeof(data);
	}
	
private:
	string symbol;
	filelink data;
};
//}}}
//{{{
class ExpressionString : public Expression
{
public:
	void create(string type, string contents)
	{
		str = contents;
	}
	void evaluate(void)
	{ }
	void write(bool dryRun)
	{
		linker.writeData( dryRun, size(), str.c_str() );
	}
	int  size(void)
	{
		return str.length() + 1;
	}
private:
	string str;
};
//}}}
//{{{
class ExpressionAlign : public Expression
{
public:
	void create(string p_type, string contents)
	{
		alignment = stringToInt(contents);
	}
	void evaluate(void)
	{
	}
	void write(bool dryRun)
	{
		linker.align(dryRun);
	}
	int  size(void)
	{
		return 0;
	}
private:
	int alignment;
};
//}}}

//{{{
int main(int argc, char **argv)
{
	int numExprs = 0;
	
	if(argc<=1)
	{
		cerr<<"Usage: mkdat [file] [-end]"<<endl;
		return 1;
	}
	if(argc>2) {
		if(!strcmp(argv[2], "-end"))
			isLittleEndian = true;
	}
	
	try
	{
		ifstream input(argv[1]);
		if(input.bad()) {
			cerr<<"Failed opening "<<argv[1]<<endl;
		}
		Array< Expression* > expressions;
		Expression *e;
		
		string attributes, type, contents;
		token_type tokenType;
		
		while(!weof(input))
		{
			type = readToken(input, tokenType);
			contents = readToken(input, tokenType);
			ws(input);
			
			if(!type.compare("ulong"))
				e = new ExpressionNumeric<ulong>;
			else if(!type.compare("slong"))
				e = new ExpressionNumeric<slong>;
			else if(!type.compare("uint"))
				e = new ExpressionNumeric<uint>;
			else if(!type.compare("sint"))
				e = new ExpressionNumeric<sint>;
			else if(!type.compare("ushort"))
				e = new ExpressionNumeric<ushort>;
			else if(!type.compare("sshort"))
				e = new ExpressionNumeric<sshort>;
			else if(!type.compare("uchar"))
				e = new ExpressionNumeric<uchar>;
			else if(!type.compare("schar"))
				e = new ExpressionNumeric<schar>;
			else if(!type.compare("link"))
				e = new ExpressionLink;
			else if(!type.compare("export"))
				e = new ExpressionExport;
			else if(!type.compare("exporttable"))
				e = new ExpressionExportTable;
			else if(!type.compare("string"))
				e = new ExpressionString;
			else if(!type.compare("align"))
				e = new ExpressionAlign;
			else
			{
				string errorMessage("Invalid type: ");
/*				errorMessage.append(type);
				errorMessage.append(" with contents ");
				errorMessage.append(contents);
				errorMessage.append(".");*/
//				errorMessage.append(numExprs);
				cerr<<numExprs<<endl;
				throw StringError(errorMessage);
			}
			expressions.append(e);
			
			expressions.end()->create(type, contents);
			
			numExprs++;
		}
		for(int i=0; i<expressions.getSize(); i++)
			expressions[i]->write(true);
		for(int i=0; i<expressions.getSize(); i++)
			expressions[i]->evaluate();
		linker.reset();
		for(int i=0; i<expressions.getSize(); i++)
			expressions[i]->write(false);
		return 0;
	}
	catch(StringError err)
	{
		cerr<<err<<endl;
		return 1;
	}
	catch(Error err)
	{}
}
//}}}

/*

Type: attributes typename name contents
typename: [name], string, ulong, slong, uint, sint, ushort, sshort, uchar, schar

*/
