%{
#include <stdio.h>
#include <string.h>
#include "mkdat.h"
 
void yyerror(const char *str)
{
	fprintf(stderr,"%s:%i: %s\n", input_filename, yylineno, str);
}
 
int yywrap()
{
	return 1;
} 
  
%}

%token TOK_IDENTIFIER TOK_TYPE TOK_NUMBER TOK_LINK TOK_EXPORT TOK_EXPORTTABLE TOK_STRING TOK_EVEN TOK_LPAREN TOK_RPAREN TOK_KEY_STRING TOK_CONST TOK_SEMI TOK_DIVERT TOK_LBRACE TOK_RBRACE
%left OP_OR
%left OP_AND
%left '|'
%left '^'
%left '&'
%left OP_EQEQ OP_NOTEQ
%left '<' '>' OP_LTEQUAL OP_GTEQUAL
%left OP_LEFTSHIFT OP_RIGHTSHIFT
%left '+' '-'
%left '*' '/' '%'
%right PREFIX    /* prefix operators (-, !, ~) */

/* TODO operators:
   ~ ?:
*/

%%

lines:
	| lines line
	;

line:
	immediate_value
	| immediate_string
	| constant_definition
	| even
	| link_reference
	| link_shuffle
	| link_export
	| link_exporttable
	| diversion
	;

const_expr:
	TOK_NUMBER                                       { $$ = $1;         }
	| TOK_IDENTIFIER                                 { $$ = get_constant($1); }
	| TOK_LPAREN const_expr TOK_RPAREN               { $$ = $2;         }
	| const_expr OP_OR const_expr                    { $$ = ($1 || $3); }
	| const_expr OP_AND const_expr                   { $$ = ($1 && $3); }
	| const_expr '|' const_expr                      { $$ = ($1 | $3);  }
	| const_expr '^' const_expr                      { $$ = $1 ^ $3;    }
	| const_expr '&' const_expr                      { $$ = $1 & $3;    }
	| const_expr OP_EQEQ const_expr                  { $$ = ($1 == $3); }
	| const_expr OP_NOTEQ const_expr                 { $$ = ($1 != $3); }
	| const_expr '<' const_expr                      { $$ = ($1 <  $3); }
	| const_expr OP_LTEQUAL const_expr               { $$ = ($1 <= $3); }
	| const_expr '>' const_expr                      { $$ = ($1 >  $3); }
	| const_expr OP_GTEQUAL const_expr               { $$ = ($1 >= $3); }
	| const_expr OP_LEFTSHIFT const_expr             { $$ = $1 << $3;   }
	| const_expr OP_RIGHTSHIFT const_expr            { $$ = $1 >> $3;   }
    | const_expr '+' const_expr                      { $$ = $1 + $3;    }
    | const_expr '-' const_expr                      { $$ = $1 - $3;    }
	| const_expr '*' const_expr                      { $$ = $1 * $3;    }
	| const_expr '/' const_expr                      { $$ = $1 / $3;    }
	| const_expr '%' const_expr                      { $$ = $1 % $3;    }
	| '-' const_expr                    %prec PREFIX { $$ = -$2;        }
	| '+' const_expr                    %prec PREFIX { $$ = +$2;        }
	| '!' const_expr                    %prec PREFIX { $$ = !($2);      }
	;

constant_definition:
	TOK_CONST TOK_IDENTIFIER '=' const_expr TOK_SEMI
	{
		define_constant($2, $4);
	}

immediate_value:
	TOK_TYPE ':' const_expr TOK_SEMI
	{
		write_val($1, $3);
	}


even:
	TOK_EVEN TOK_SEMI
	{
		write_even();
	}

immediate_string:
	TOK_KEY_STRING ':' TOK_STRING TOK_SEMI
	{
		write_string($3);
	}

link_reference:
	TOK_LINK ':' TOK_IDENTIFIER TOK_SEMI
	{
		write_reference($3);
	}

link_shuffle:
	TOK_LINK ':' TOK_LPAREN const_expr TOK_RPAREN TOK_SEMI
	{
		write_shuffle_reference($4);
	}

link_export:
	TOK_EXPORT TOK_IDENTIFIER TOK_SEMI
	{
		write_label($2, 0);
	}

link_exporttable:
	TOK_EXPORTTABLE TOK_IDENTIFIER TOK_SEMI
	{
		write_label($2, 1);
	}

diversion:
	TOK_DIVERT TOK_LBRACE
	{
		start_diversion();
	} const_expr TOK_RBRACE
	{
		finish_diversion();
	}

